#include <complex>
#include <iostream>
#include <cerrno>
#include <csignal>

#include <cadef.h>
#include <epicsVersion.h>

#include "mdb.h"
#include "scan.h"
#include "SDDS.h"
#include "SDDSepics.h"
#include "pvaSDDS.h"

//#include "../modules/pvAccess/src/ca/caChannel.h"

#define CLO_SET_PIPE 0
#define CLO_VERBOSE 1
#define CLO_PENDIOTIME 2
#define COMMANDLINE_OPTIONS 3

char *commandline_option[COMMANDLINE_OPTIONS] = {
  (char *)"pipe",
  (char *)"verbose",
  (char *)"pendIOTime",
};

char *USAGE = (char *)"sddscainfo [<inputfile>] [<outputfile>] \n\
 [-pipe=[input][,output]] \n\
 [-verbose] [-pendIOTime=<seconds>]\n\
\n\
Required SDDS columns: ControlName (string)\n\
Program by Robert Soliday, ANL\n\
Link date: " __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ", " EPICS_VERSION_STRING "\n";

typedef struct
{
  int64_t pvCount;
  char **ControlName;
} SDDS_INPUT_DATA_PAGE;

typedef struct
{
  int64_t pvCount;
  char **ControlName;
  char *Active;
  char **Provider;
  char **IOCHost;
  char *ReadAccess;
  char *WriteAccess;
  char **StructureID;
  char **FieldType;
  uint32_t *ElementCount;
  char **NativeDataType;
  char **Units;
  uint32_t *EnumNumOfChoices;
  char ***EnumChoices;
} SDDS_OUTPUT_DATA;

typedef struct
{
  bool verbose;
  char *inputFile;
  char *outputFile;
  unsigned long pipeFlags;
  long tmpfile_used;
  double pendIOTime;
  SDDS_INPUT_DATA_PAGE *inputPage;
  SDDS_OUTPUT_DATA outputData;
  int32_t inputPages;
  int64_t total_pvCount;
} PROG_DATA;

int ReadArguments(int argc, char **argv, PROG_DATA *prog_data);
int ReadInput(PROG_DATA *prog_data);
int ConnectPVs(PVA_OVERALL *pva, PVA_OVERALL *pva_ca, PROG_DATA *prog_data);
int GetPVinfo(PVA_OVERALL *pva_pva, PVA_OVERALL *pva_ca, PROG_DATA *prog_data);
int WriteOutput(PROG_DATA *prog_data);
void FreeMemory(PROG_DATA *prog_data);
long pvaThreadSleep(long double seconds);

#if defined(_WIN32)
#  include <thread>
#  include <chrono>
int nanosleep(const struct timespec *req, struct timespec *rem) {
  long double ldSeconds;
  long mSeconds;
  ldSeconds = req->tv_sec + req->tv_nsec;
  mSeconds = ldSeconds / 1000000;
  std::this_thread::sleep_for(std::chrono::microseconds(mSeconds));
  return (0);
}
#endif

int main(int argc, char **argv) {
  PVA_OVERALL pva, pva_ca;
  PROG_DATA prog_data;

  if (ReadArguments(argc, argv, &prog_data) == 1) {
    return (1);
  }
  if (ReadInput(&prog_data) == 1) {
    return (1);
  }
  if (ConnectPVs(&pva, &pva_ca, &prog_data) == 1) {
    return (1);
  }
  if (ExtractPVAUnits(&pva) == 1) {
    return (1);
  }
  if (ExtractPVAUnits(&pva_ca) == 1) {
    return (1);
  }
  if (GetPVinfo(&pva, &pva_ca, &prog_data) == 1) {
    return (1);
  }
  if (WriteOutput(&prog_data) == 1) {
    return (1);
  }
  FreeMemory(&prog_data);

  freePVA(&pva);
  freePVA(&pva_ca);
  return (0);
}

int ReadArguments(int argc, char **argv, PROG_DATA *prog_data) {
  SCANNED_ARG *s_arg;
  long i_arg;

  prog_data->verbose = false;
  prog_data->inputFile = NULL;
  prog_data->outputFile = NULL;
  prog_data->pipeFlags = 0;
  prog_data->pendIOTime = .5;

  SDDS_RegisterProgramName("sddscainfo");

  argc = scanargs(&s_arg, argc, argv);
  if (argc < 2) {
    fprintf(stderr, "%s\n", USAGE);
    return (1);
  }
  for (i_arg = 1; i_arg < argc; i_arg++) {
    if (s_arg[i_arg].arg_type == OPTION) {
      switch (match_string(s_arg[i_arg].list[0], commandline_option, COMMANDLINE_OPTIONS, 0)) {
      case CLO_SET_PIPE:
        if (!processPipeOption(s_arg[i_arg].list + 1, s_arg[i_arg].n_items - 1, &prog_data->pipeFlags)) {
          fprintf(stderr, "invalid -pipe syntax\n");
          return (1);
        }
        break;
      case CLO_PENDIOTIME:
        if (s_arg[i_arg].n_items != 2 ||
            !(get_double(&prog_data->pendIOTime, s_arg[i_arg].list[1]))) {
          fprintf(stderr, "no value given for option -pendIOTime\n");
          return (1);
        }
        break;
      case CLO_VERBOSE:
        prog_data->verbose = true;
        break;
      }
    } else {
      if (!prog_data->inputFile)
        SDDS_CopyString(&prog_data->inputFile, s_arg[i_arg].list[0]);
      else if (!prog_data->outputFile)
        SDDS_CopyString(&prog_data->outputFile, s_arg[i_arg].list[0]);
      else {
        fprintf(stderr, "too many filenames given\n");
        return (1);
      }
    }
  }
  if ((prog_data->pipeFlags == 0) && (prog_data->outputFile == NULL)) {
    fprintf(stderr, "%s\n", USAGE);
    return (1);
  }

  prog_data->tmpfile_used = 0;
  processFilenames((char*)"sddscainfo", &prog_data->inputFile, &prog_data->outputFile, prog_data->pipeFlags, 1, &prog_data->tmpfile_used);

  free_scanargs(&s_arg, argc);
  return (0);
}


int ReadInput(PROG_DATA *prog_data) {
  SDDS_DATASET input;
  prog_data->inputPages = 0;
  prog_data->total_pvCount = 0;
  int32_t page=0;

  if (!SDDS_InitializeInput(&input, prog_data->inputFile)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return (1);
  }
  if (SDDS_CheckColumn(&input, (char *)"ControlName", NULL, SDDS_STRING, NULL) != SDDS_CHECK_OK) {
    fprintf(stderr, "Column \"ControlName\" does not exist or something wrong with it in input file %s\n", prog_data->inputFile);
    return (1);
  }
  while (SDDS_ReadPage(&input) > 0) {
    page++;
    if (page == 1) {
      prog_data->inputPage = (SDDS_INPUT_DATA_PAGE *)malloc(sizeof(SDDS_INPUT_DATA_PAGE));
    } else {
      prog_data->inputPage = (SDDS_INPUT_DATA_PAGE *)realloc(prog_data->inputPage, sizeof(SDDS_INPUT_DATA_PAGE) * page);
    }
    prog_data->inputPages = page;
    prog_data->inputPage[page-1].pvCount = input.n_rows;
    prog_data->total_pvCount += input.n_rows;
    prog_data->inputPage[page-1].ControlName = NULL;
    if (input.n_rows > 0) {
      if (!(prog_data->inputPage[page-1].ControlName = (char **)SDDS_GetColumn(&input, (char *)"ControlName"))) {
        fprintf(stderr, (char *)"No data provided in the input file.\n");
        return (1);
      }
    }
  }
  if (!SDDS_Terminate(&input)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return (1);
  }
  return (0);
}

int WriteOutput(PROG_DATA *prog_data) {
  int64_t i;
  SDDS_DATASET output;
  char **EnumChoices;

  if (SDDS_InitializeOutput(&output, SDDS_BINARY, 1,
                            NULL, NULL, prog_data->outputFile) != 1) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return(1);
  }
  output.layout.data_mode.column_major = SDDS_COLUMN_MAJOR_ORDER;

  if (SDDS_DefineSimpleColumn(&output, "ControlName",      NULL, SDDS_STRING)!=1 ||
      SDDS_DefineSimpleColumn(&output, "Active",           NULL, SDDS_CHARACTER)!=1 ||
      SDDS_DefineSimpleColumn(&output, "Provider",         NULL, SDDS_STRING)!=1 ||
      SDDS_DefineSimpleColumn(&output, "IOCHost",          NULL, SDDS_STRING)!=1 ||
      SDDS_DefineSimpleColumn(&output, "ReadAccess",       NULL, SDDS_CHARACTER)!=1 ||
      SDDS_DefineSimpleColumn(&output, "WriteAccess",      NULL, SDDS_CHARACTER)!=1 ||
      SDDS_DefineSimpleColumn(&output, "StructureID",      NULL, SDDS_STRING)!=1 ||
      SDDS_DefineSimpleColumn(&output, "FieldType",        NULL, SDDS_STRING)!=1 ||
      SDDS_DefineSimpleColumn(&output, "ElementCount",     NULL, SDDS_ULONG)!=1 ||
      SDDS_DefineSimpleColumn(&output, "NativeDataType",   NULL, SDDS_STRING)!=1 ||
      SDDS_DefineSimpleColumn(&output, "Units",            NULL, SDDS_STRING)!=1 ||
      SDDS_DefineSimpleColumn(&output, "EnumChoices",      NULL, SDDS_STRING)!=1) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return(1);
  }
  EnumChoices = (char**)malloc(sizeof(char*) * prog_data->outputData.pvCount);
  for (i=0; i<prog_data->outputData.pvCount; i++) {
    if (prog_data->outputData.EnumNumOfChoices[i] > 0) {
      int length = 0, index = 0;
      for (uint32_t j = 0; j < prog_data->outputData.EnumNumOfChoices[i]; j++) {
        length += strlen(prog_data->outputData.EnumChoices[i][j]) + 1;
      }
      EnumChoices[i] = (char*)malloc(sizeof(char) * length);
      for (uint32_t j = 0; j < prog_data->outputData.EnumNumOfChoices[i]; j++) {
        strcpy(&EnumChoices[i][index], prog_data->outputData.EnumChoices[i][j]);
        index += strlen(prog_data->outputData.EnumChoices[i][j]);
        EnumChoices[i][index] = ',';
        index++;
      }
      EnumChoices[i][length - 1] = '\0';
    } else {
      SDDS_CopyString(&EnumChoices[i], "");
    }
  }

  if (SDDS_WriteLayout(&output)!=1) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return(1);
  }
  if (SDDS_StartPage(&output, prog_data->outputData.pvCount)!=1) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return(1);
  }

  if (SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.ControlName,      prog_data->outputData.pvCount, "ControlName")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.Active,           prog_data->outputData.pvCount, "Active")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.Provider,         prog_data->outputData.pvCount, "Provider")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.IOCHost,          prog_data->outputData.pvCount, "IOCHost")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.ReadAccess,       prog_data->outputData.pvCount, "ReadAccess")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.WriteAccess,      prog_data->outputData.pvCount, "WriteAccess")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.StructureID,      prog_data->outputData.pvCount, "StructureID")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.FieldType,        prog_data->outputData.pvCount, "FieldType")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.ElementCount,     prog_data->outputData.pvCount, "ElementCount")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.NativeDataType,   prog_data->outputData.pvCount, "NativeDataType")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, prog_data->outputData.Units,            prog_data->outputData.pvCount, "Units")!=1 ||
      SDDS_SetColumn(&output, SDDS_SET_BY_NAME, EnumChoices,                            prog_data->outputData.pvCount, "EnumChoices")!=1) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return(1);
  }

  if (SDDS_WritePage(&output)!=1) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return(1);
  }
  if (SDDS_Terminate(&output)!=1) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    return(1);
  }
  if (prog_data->tmpfile_used && !replaceFileAndBackUp(prog_data->inputFile, prog_data->outputFile))
    exit(1);

  for (int64_t j=0; j<prog_data->outputData.pvCount; j++) {
    if (EnumChoices[j])
      free(EnumChoices[j]);
  }
  free(EnumChoices);

  return (0);
}

int ConnectPVs(PVA_OVERALL *pva, PVA_OVERALL *pva_ca, PROG_DATA *prog_data) {
  long i, j, p;
  //Allocate memory for pva structures
  allocPVA(pva, prog_data->total_pvCount);
  allocPVA(pva_ca, prog_data->total_pvCount);
  //List PV names
  epics::pvData::shared_vector<std::string> names(pva->numPVs);
  epics::pvData::shared_vector<std::string> providerNames(pva->numPVs);
  epics::pvData::shared_vector<std::string> providerNames_ca(pva->numPVs);
  j = 0;
  for (p = 0; p < prog_data->inputPages; p++) {
    for (i = 0; i < prog_data->inputPage[p].pvCount; i++) {
      names[j] = prog_data->inputPage[p].ControlName[i];
      providerNames[j] = "pva";
      providerNames_ca[j] = "ca";
      j++;
    }
  }
  pva->pvaChannelNames = pva_ca->pvaChannelNames = freeze(names);
  pva->pvaProvider = freeze(providerNames);
  pva_ca->pvaProvider = freeze(providerNames_ca);

  ConnectPVA(pva, prog_data->pendIOTime);
  if (pvaThreadSleep(.1) == 1) {
    return (1);
  }

  ConnectPVA(pva_ca, prog_data->pendIOTime);
  if (pvaThreadSleep(.1) == 1) {
    return (1);
  }
  if (GetPVAValues(pva) == 1) {
    return (1);
  }
  if (GetPVAValues(pva_ca) == 1) {
    return (1);
  }
  return (0);
}

int GetPVinfo(PVA_OVERALL *pva_pva, PVA_OVERALL *pva_ca, PROG_DATA *prog_data) {
  long i, j, p;
  PVA_OVERALL *pva=NULL;

  prog_data->outputData.pvCount = prog_data->total_pvCount;
  prog_data->outputData.ControlName    =    (char**)malloc(sizeof(char*)    * prog_data->outputData.pvCount);
  prog_data->outputData.Active         =    (char *)malloc(sizeof(char)     * prog_data->outputData.pvCount);
  prog_data->outputData.Provider       =    (char**)malloc(sizeof(char*)    * prog_data->outputData.pvCount);
  prog_data->outputData.IOCHost        =    (char**)malloc(sizeof(char*)    * prog_data->outputData.pvCount);
  prog_data->outputData.ReadAccess     =    (char *)malloc(sizeof(char)     * prog_data->outputData.pvCount);
  prog_data->outputData.WriteAccess    =    (char *)malloc(sizeof(char)     * prog_data->outputData.pvCount);
  prog_data->outputData.StructureID    =    (char**)malloc(sizeof(char*)    * prog_data->outputData.pvCount);
  prog_data->outputData.FieldType      =    (char**)malloc(sizeof(char*)    * prog_data->outputData.pvCount);
  prog_data->outputData.ElementCount   = (uint32_t*)malloc(sizeof(uint32_t) * prog_data->outputData.pvCount);
  prog_data->outputData.NativeDataType =    (char**)malloc(sizeof(char*)    * prog_data->outputData.pvCount);
  prog_data->outputData.Units          =    (char**)malloc(sizeof(char*)    * prog_data->outputData.pvCount);
  prog_data->outputData.EnumNumOfChoices=(uint32_t*)malloc(sizeof(uint32_t) * prog_data->outputData.pvCount);
  prog_data->outputData.EnumChoices    =   (char***)malloc(sizeof(char**)   * prog_data->outputData.pvCount);

  j = 0;
  for (p = 0; p < prog_data->inputPages; p++) {
    for (i = 0; i < prog_data->inputPage[p].pvCount; i++) {
      prog_data->outputData.EnumNumOfChoices[j] = 0;
      SDDS_CopyString(&prog_data->outputData.ControlName[j], prog_data->inputPage[p].ControlName[i]);
      if (pva_pva->isConnected[j])
        pva = pva_pva;
      else
        pva = pva_ca;

      if (pva->isConnected[j])
        prog_data->outputData.Active[j] = 'y';
      else
        prog_data->outputData.Active[j] = 'n';
      SDDS_CopyString(&prog_data->outputData.Provider[j], GetProviderName(pva, j).c_str());
      SDDS_CopyString(&prog_data->outputData.IOCHost[j], GetRemoteAddress(pva,j).c_str());
      if (HaveReadAccess(pva, j))
        prog_data->outputData.ReadAccess[j] = 'y';
      else
        prog_data->outputData.ReadAccess[j] = 'n';
      if (HaveWriteAccess(pva, j))
        prog_data->outputData.WriteAccess[j] = 'y';
      else
        prog_data->outputData.WriteAccess[j] = 'n';
      SDDS_CopyString(&prog_data->outputData.StructureID[j], GetStructureID(pva, j).c_str());
      SDDS_CopyString(&prog_data->outputData.FieldType[j], GetFieldType(pva,j).c_str());
      prog_data->outputData.ElementCount[j] = GetElementCount(pva,j);
      SDDS_CopyString(&prog_data->outputData.NativeDataType[j], GetNativeDataType(pva,j).c_str());
      SDDS_CopyString(&prog_data->outputData.Units[j], GetUnits(pva,j).c_str());
      if (IsEnumFieldType(pva,j)) {
        prog_data->outputData.EnumNumOfChoices[j] = GetEnumChoices(pva, j, &prog_data->outputData.EnumChoices[j]);
      }
      
      if (prog_data->verbose) {
        fprintf(stdout, "%s\n", prog_data->outputData.ControlName[j]);
        fprintf(stdout, "            Active:   %c\n", prog_data->outputData.Active[j]);
        fprintf(stdout, "          Provider:   %s\n", prog_data->outputData.Provider[j]);
        fprintf(stdout, "              Host:   %s\n", prog_data->outputData.IOCHost[j]);
        fprintf(stdout, "       Read Access:   %c\n", prog_data->outputData.ReadAccess[j]);
        fprintf(stdout, "      Write Access:   %c\n", prog_data->outputData.WriteAccess[j]);
        fprintf(stdout, "      Structure ID:   %s\n", prog_data->outputData.StructureID[j]);
        fprintf(stdout, "        Field Type:   %s\n", prog_data->outputData.FieldType[j]);
        fprintf(stdout, "     Element Count:   %" PRIu32 "\n", prog_data->outputData.ElementCount[j]);
        fprintf(stdout, "  Native Data Type:   %s\n", prog_data->outputData.NativeDataType[j]);
        fprintf(stdout, "             Units:   %s\n", prog_data->outputData.Units[j]);
        if (IsEnumFieldType(pva,j)) {
          fprintf(stdout, " Number of Choices:   %" PRIu32 "\n", prog_data->outputData.EnumNumOfChoices[j]);
          for (uint32_t m = 0; m < prog_data->outputData.EnumNumOfChoices[j]; m++) {
            fprintf(stdout, "           Index %" PRIu32 ":   %s\n", m, prog_data->outputData.EnumChoices[j][m]);
          }
        }
        fprintf(stdout, "\n");
      }
      j++;
    }
  }
  return (0);
}

void FreeMemory(PROG_DATA *prog_data) {
  for (int32_t i=0; i<prog_data->inputPages; i++) {
    if (prog_data->inputPage[i].ControlName) {
      for (int64_t j=0; j<prog_data->inputPage[i].pvCount; j++) {
        if (prog_data->inputPage[i].ControlName[j])
          free(prog_data->inputPage[i].ControlName[j]);
      }
      free (prog_data->inputPage[i].ControlName);
    }
  }
  free(prog_data->inputPage);
  for (int64_t i=0; i<prog_data->outputData.pvCount; i++) {
    if (prog_data->outputData.ControlName[i])
      free(prog_data->outputData.ControlName[i]);
    if (prog_data->outputData.Provider[i])
      free(prog_data->outputData.Provider[i]);
    if (prog_data->outputData.IOCHost[i])
      free(prog_data->outputData.IOCHost[i]);
    if (prog_data->outputData.StructureID[i])
      free(prog_data->outputData.StructureID[i]);
    if (prog_data->outputData.FieldType[i])
      free(prog_data->outputData.FieldType[i]);
    if (prog_data->outputData.NativeDataType[i])
      free(prog_data->outputData.NativeDataType[i]);
    if (prog_data->outputData.Units[i])
      free(prog_data->outputData.Units[i]);
    if (prog_data->outputData.EnumNumOfChoices[i] > 0) {
      if (prog_data->outputData.EnumChoices[i]) {
        for (uint32_t j=0; j<prog_data->outputData.EnumNumOfChoices[i]; j++) {
          if (prog_data->outputData.EnumChoices[i][j])
            free(prog_data->outputData.EnumChoices[i][j]);
        }
        free(prog_data->outputData.EnumChoices[i]);
      }
    }

  }
  free(prog_data->outputData.ControlName);
  free(prog_data->outputData.Active);
  free(prog_data->outputData.Provider);
  free(prog_data->outputData.IOCHost);
  free(prog_data->outputData.ReadAccess);
  free(prog_data->outputData.WriteAccess);
  free(prog_data->outputData.StructureID);
  free(prog_data->outputData.FieldType);
  free(prog_data->outputData.ElementCount);
  free(prog_data->outputData.NativeDataType);
  free(prog_data->outputData.Units);
  free(prog_data->outputData.EnumNumOfChoices);
  free(prog_data->outputData.EnumChoices);
  if (prog_data->inputFile)
    free(prog_data->inputFile);
  if (prog_data->outputFile)
    free(prog_data->outputFile);
}

long pvaThreadSleep(long double seconds) {
  struct timespec delayTime;
  struct timespec remainingTime;

  if (seconds > 0) {
    delayTime.tv_sec = seconds;
    delayTime.tv_nsec = (seconds - delayTime.tv_sec) * 1e9;
    while (nanosleep(&delayTime, &remainingTime) == -1 &&
           errno == EINTR)
      delayTime = remainingTime;
  }
  return (0);
}
